import copy
import logging
from typing import Tuple, Union

from spaceone.core.service import *
from spaceone.board.error import *

from spaceone.board.manager.post_manager import PostManager
from spaceone.board.manager.file_manager import FileManager
from spaceone.board.model.post_model import Post

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class PostService(BaseService):
    resource = "Post"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.post_mgr: PostManager = self.locator.get_manager(PostManager)
        self.file_mgr = None

    @transaction(
        permission="board:Post.write",
        role_types=["SYSTEM_ADMIN", "DOMAIN_ADMIN"],
    )
    @check_required(["board_type", "title", "contents"])
    def create(self, params: dict) -> Post:
        """Create post

        Args:
            params (dict): {
                'board_type': 'str',       # required
                'category': 'str',
                'title': 'str',            # required
                'contents': 'str',         # required
                'files' : 'list',
                'options': 'dict',
                'writer': 'str',
                'resource_group': 'str',   # required
                'domain_id': 'str',        # injected from auth
                'user_id': 'str'           # injected from auth
            }

        Returns:
            post_vo (object)
        """
        resource_group = params["resource_group"]
        if resource_group == "SYSTEM":
            params["domain_id"] = "*"

        _options = {"is_pinned": False, "is_popup": False}
        if options := params.get("options", {}):
            self._valid_options(options)

            if "is_pinned" in options:
                _options.update({"is_pinned": options["is_pinned"]})
            if "is_popup" in options:
                _options.update({"is_popup": options["is_popup"]})
        params["options"] = _options

        file_ids = params.get("files", [])
        self.file_mgr: FileManager = self.locator.get_manager(FileManager)
        self._check_files(file_ids)
        post_vo = self.post_mgr.create_post(params)

        self._update_file_reference(post_vo.post_id, file_ids)

        return post_vo

    @transaction(
        permission="board:Post.write",
        role_types=["SYSTEM_ADMIN", "DOMAIN_ADMIN"],
    )
    @check_required(["post_id"])
    def update(self, params: dict) -> Post:
        """Update post

        Args:
            params (dict): {
                'post_id': 'str',    # required
                'category': 'str',
                'title': 'str',
                'contents': 'str',
                'files' : 'list',
                'options': 'dict',
                'writer': 'str',
                'domain_id': 'str',        # injected from auth
                'user_id': 'str'           # injected from auth
            }

        Returns:
            post_vo (object)
        """
        post_id = params["post_id"]
        domain_id = params.get("domain_id")

        post_vo = self.post_mgr.get_post(post_id, domain_id)

        if options := params.get("options", {}):
            self._valid_options(options)

            _options = copy.deepcopy(post_vo.options)
            _options.update(options)

            params["options"] = _options

        if "files" in params:
            self.file_mgr: FileManager = self.locator.get_manager(FileManager)

            new_file_ids = set(params["files"])
            old_file_ids = set(post_vo.files)

            if new_file_ids != old_file_ids:
                file_ids_to_be_deleted = list(old_file_ids - new_file_ids)
                file_ids_to_be_created = list(new_file_ids - old_file_ids)

                if len(file_ids_to_be_created) > 0:
                    self._check_files(file_ids_to_be_created)
                    self._update_file_reference(post_vo.post_id, file_ids_to_be_created)

                if len(file_ids_to_be_deleted) > 0:
                    for file_id in file_ids_to_be_deleted:
                        self.file_mgr.delete_file(file_id)

        return self.post_mgr.update_post_by_vo(params, post_vo)

    @transaction(
        permission="board:Post.write",
        role_types=["SYSTEM_ADMIN", "DOMAIN_ADMIN"],
    )
    @check_required(["post_id"])
    def send_notification(self, params: dict) -> None:
        """Delete post

        Args:
            params (dict): {
                'post_id': 'str',     # required
                'domain_id': 'str'    # injected from auth
                'user_id': 'str'      # injected from auth
            }

        Returns:
            None
        """
        pass

    @transaction(
        permission="board:Post.write",
        role_types=["SYSTEM_ADMIN", "DOMAIN_ADMIN"],
    )
    @check_required(["post_id"])
    def delete(self, params: dict) -> None:
        """Delete post

        Args:
            params (dict): {
                'post_id': 'str',     # required
                'domain_id': 'str'    # injected from auth
                'user_id': 'str'      # injected from auth
            }

        Returns:
            None
        """

        post_id = params["post_id"]
        domain_id = params.get("domain_id")

        post_vo = self.post_mgr.get_post(post_id, domain_id)

        if len(post_vo.files) > 0:
            self.file_mgr: FileManager = self.locator.get_manager(FileManager)

            for file_id in post_vo.files:
                self.file_mgr.delete_file(file_id)

        self.post_mgr.delete_post_vo(post_vo)

    @transaction(
        permission="board:Post.read",
        role_types=[
            "SYSTEM_ADMIN",
            "DOMAIN_ADMIN",
            "WORKSPACE_OWNER",
            "WORKSPACE_MEMBER",
        ],
    )
    @change_value_by_rule("APPEND", "domain_id", "*")
    @check_required(["post_id"])
    def get(self, params: dict) -> Tuple[Post, list]:
        """Get post

        Args:
            params (dict): {
                'post_id': 'str',     # required
                'domain_id': 'str,    # injected from auth
                'user_id': 'str'      # injected from auth
            }

        Returns:
            post_vo (object)
        """

        post_id = params["post_id"]
        domain_id = params.get("domain_id")

        post_vo = self.post_mgr.get_post(post_id, domain_id)
        self.post_mgr.increase_view_count(post_vo)

        self.file_mgr: FileManager = self.locator.get_manager(FileManager)

        files_info = []
        if len(post_vo.files) > 0:
            files_info = self._get_files_info_from_file_manager(post_vo.files)

        return post_vo, files_info

    @transaction(
        permission="board:Post.read",
        role_types=[
            "SYSTEM_ADMIN",
            "DOMAIN_ADMIN",
            "WORKSPACE_OWNER",
            "WORKSPACE_MEMBER",
        ],
    )
    @change_value_by_rule("APPEND", "domain_id", "*")
    @append_query_filter(
        [
            "board_type",
            "post_id",
            "category",
            "writer",
            "domain_id"
        ]
    )
    def list(self, params: dict) -> dict:
        """List posts

        Args:
            params (dict): {
                'query': 'dict',
                'board_type': 'str',
                'post_id': 'str',
                'category': 'str',
                'writer': 'str',
                'is_pinned': 'bool',
                'is_popup': 'bool',
                'domain_id': 'str',        # injected from auth
                'workspace_id': 'str',     # injected from auth
                'user_projects': 'list'    # injected from auth
                'user_id': 'str'           # injected from auth
            }

        Returns:
            posts_vo (object)
            total_count
        """

        query = params.get("query", {})
        return self.post_mgr.list_boards(query)

    @transaction(
        permission="board:Post.read",
        role_types=[
            "SYSTEM_ADMIN",
            "DOMAIN_ADMIN",
            "WORKSPACE_OWNER",
            "WORKSPACE_MEMBER",
        ],
    )
    @change_value_by_rule("APPEND", "domain_id", "*")
    @check_required(["query"])
    @append_query_filter(["domain_id"])
    def stat(self, params: dict) -> dict:
        """List posts

        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)',    # required
                'domain_id': 'str',                                        # injected from auth
                'workspace_id': 'str',                                     # injected from auth
                'user_projects': 'list'                                    # injected from auth
                'user_id': 'str'                                           # injected from auth
            }

        Returns:
            dict: {
                'results': 'list',
                'total_count': 'int'
            }
        """

        query = params.get("query", {})
        return self.post_mgr.stat_boards(query)

    def _check_files(self, file_ids: list) -> None:
        for file_id in file_ids:
            self._verify_file(file_id)

    def _update_file_reference(self, post_id: str, files: list) -> None:
        for file in files:
            reference = {"resource_id": post_id, "resource_type": "board.Post"}
            self.file_mgr.update_file_reference(file, reference)

    def _verify_file(self, file_id: str) -> None:
        file_info = self.file_mgr.get_file(file_id)

        file_state = file_info.get("state")
        if file_state != "DONE":
            raise ERROR_INVALID_FILE_STATE(file_id=file_id, state=file_state)

    def _get_files_info_from_file_manager(self, file_ids: list):
        files_info = []
        for file_id in file_ids:
            file_info = self.file_mgr.get_download_url(file_id)
            files_info.append(file_info)
        return files_info

    @staticmethod
    def _valid_options(options: dict) -> None:
        exact_keys = ["is_pinned", "is_popup"]

        for key in options:
            if key not in exact_keys:
                raise ERROR_INVALID_KEY_IN_OPTIONS



import functools

from spaceone.api.board.v1 import post_pb2
from spaceone.core.pygrpc.message_type import change_struct_type, change_list_value_type
from spaceone.core import utils

from spaceone.board.model.post_model import Post

__all__ = ["PostInfo", "PostsInfo"]


def PostInfo(post_vo: Post, minimal=False, files_info=None):
    info = {
        "board_type": post_vo.board_type,
        "post_id": post_vo.post_id,
        "category": post_vo.category,
        "title": post_vo.title,
        "resource_group": post_vo.resource_group,
        "domain_id": post_vo.domain_id,
    }

    if not minimal:
        info.update(
            {
                "contents": post_vo.contents,
                "view_count": post_vo.view_count,
                "writer": post_vo.writer,
                "options": change_struct_type(post_vo.options),
                "user_id": post_vo.user_id,
                "created_at": utils.datetime_to_iso8601(post_vo.created_at),
                "updated_at": utils.datetime_to_iso8601(post_vo.updated_at),
            }
        )

        if files_info:
            info["files"] = change_list_value_type(files_info)
        else:
            info["files"] = change_list_value_type(post_vo.files)

    return post_pb2.PostInfo(**info)


def PostsInfo(post_vos, total_count, **kwargs):
    return post_pb2.PostsInfo(
        results=list(map(functools.partial(PostInfo, **kwargs), post_vos)),
        total_count=total_count,
    )



# Database Settings
DATABASES = {
    "default": {
        "db": "board",
        "host": "localhost",
        "port": 27017,
        "username": "",
        "password": "",
    }
}

# Cache Settings
CACHES = {
    "default": {
        # Redis Example
        # 'backend': 'spaceone.core.cache.redis_cache.RedisCache',
        # 'host': '<host>',
        # 'port': 6379,
        # 'db': 0
    }
}

# Handler Configuration
HANDLERS = {
    "authentication": [
        # Default Authentication Handler
        # {
        #     'backend': 'spaceone.core.handler.authentication_handler.AuthenticationGRPCHandler',
        #     'uri': 'grpc://identity:50051/v1/Domain/get_public_key'
        # }
    ],
    "authorization": [
        # Default Authorization Handler
        # {
        #     'backend': 'spaceone.core.handler.authorization_handler.AuthorizationGRPCHandler',
        #     'uri': 'grpc://identity:50051/v1/Authorization/verify'
        # }
    ],
    "mutation": [],
    "event": [],
}

# Connector Settings
CONNECTORS = {
    "SpaceConnector": {
        "backend": "spaceone.core.connector.space_connector:SpaceConnector",
        "endpoints": {
            "identity": "grpc://identity:50051",
            "file_manager": "grpc://file-manager:50051",
        },
    }
}

# Log Settings
LOG = {}



from spaceone.api.board.v1 import post_pb2, post_pb2_grpc
from spaceone.core.pygrpc import BaseAPI
from spaceone.board.service.post_service import PostService
from spaceone.board.info.post_info import *
from spaceone.board.info.common_info import *


class Post(BaseAPI, post_pb2_grpc.PostServicer):
    pb2 = post_pb2
    pb2_grpc = post_pb2_grpc

    def create(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service(PostService, metadata) as post_service:
            return self.locator.get_info(PostInfo, post_service.create(params))

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service(PostService, metadata) as post_service:
            return self.locator.get_info(PostInfo, post_service.update(params))

    def delete(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service(PostService, metadata) as post_service:
            post_service.delete(params)
            return self.locator.get_info(EmptyInfo)

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service(PostService, metadata) as post_service:
            post_vo, files_info = post_service.get(params)
            return self.locator.get_info(PostInfo, post_vo, files_info=files_info)

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service(PostService, metadata) as post_service:
            post_vos, total_count = post_service.list(params)
            return self.locator.get_info(
                PostsInfo, post_vos, total_count, minimal=self.get_minimal(params)
            )

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service(PostService, metadata) as post_service:
            return self.locator.get_info(StatisticsInfo, post_service.stat(params))



import logging

from spaceone.core import cache
from spaceone.core.manager import BaseManager
from spaceone.core.connector.space_connector import SpaceConnector

_LOGGER = logging.getLogger(__name__)


class FileManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.file_manager_connector: SpaceConnector = self.locator.get_connector(
            SpaceConnector, service="file_manager"
        )

    def get_file(self, file_id: str) -> dict:
        return self.file_manager_connector.dispatch("File.get", {"file_id": file_id})

    def update_file_reference(self, file_id: str, reference: dict) -> dict:
        return self.file_manager_connector.dispatch(
            "File.update",
            {"file_id": file_id, "reference": reference},
        )

    def delete_file(self, file_id: str) -> None:
        return self.file_manager_connector.dispatch("File.delete", {"file_id": file_id})

    @cache.cacheable(key="board:file-download-url:{file_id}", expire=600)
    def get_download_url(self, file_id: str) -> dict:
        return self.file_manager_connector.dispatch(
            "File.get_download_url", {"file_id": file_id}
        )



from spaceone.core.error import *


class ERROR_INVALID_KEY_IN_OPTIONS(ERROR_BASE):
    _message = "Invalid key in options"


class ERROR_INVALID_FILE_STATE(ERROR_BASE):
    _message = "file state is invalid (file_id = {file_id}, state = {state})"



import logging
from typing import Union

from spaceone.core.manager import BaseManager
from spaceone.board.model.post_model import Post

_LOGGER = logging.getLogger(__name__)


class PostManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.post_model: Post = self.locator.get_model(Post)

    def create_post(self, params: dict) -> Post:
        def _rollback(vo: Post):
            _LOGGER.info(f"[create_post._rollback] " f"Delete post : {vo.name}")
            vo.delete()

        post_vo: Post = self.post_model.create(params)
        self.transaction.add_rollback(_rollback, post_vo)

        return post_vo

    def update_post(self, params: dict) -> Post:
        post_vo: Post = self.get_post(params["board_id"], params["post_id"])
        return self.update_post_by_vo(params, post_vo)

    def update_post_by_vo(self, params: dict, post_vo: Post) -> Post:
        def _rollback(old_data: dict):
            _LOGGER.info(
                f"[update_post_by_vo._rollback] Revert Data : " f'{old_data["post_id"]}'
            )
            post_vo.update(old_data)

        self.transaction.add_rollback(_rollback, post_vo.to_dict())

        return post_vo.update(params)

    def get_post(self, post_id: str, domain_id: str = None) -> Post:
        conditions = {"post_id": post_id}

        if domain_id:
            conditions["domain_id"] = domain_id

        return self.post_model.get(**conditions)

    def list_boards(self, query: Union[dict, None] = None):
        if query is None:
            query = {}
        return self.post_model.query(**query)

    def stat_boards(self, query: dict) -> dict:
        return self.post_model.stat(**query)

    @staticmethod
    def increase_view_count(post_vo: Post) -> None:
        post_vo.increment("view_count")

    @staticmethod
    def delete_post_vo(post_vo: Post) -> None:
        post_vo.delete()



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel


class Post(MongoModel):
    board_type = StringField(max_length=40)
    post_id = StringField(max_length=40, generate_id="post", unique=True)
    category = StringField(null=True, default=None)
    title = StringField(max_length=255)
    contents = StringField()
    options = DictField(default={})
    view_count = IntField(default=0)
    writer = StringField()
    files = ListField(StringField(), default=[])
    resource_group = StringField(max_length=40, choices=("SYSTEM", "DOMAIN"))
    domain_id = StringField(max_length=40)
    user_id = StringField(max_length=40)
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)

    meta = {
        "updatable_fields": [
            "category",
            "title",
            "contents",
            "options",
            "writer",
            "files",
        ],
        "minimal_fields": [
            "board_type",
            "post_id",
            "category",
            "title",
            "resource_group",
            "domain_id",
        ],
        "ordering": ["-created_at"],
        "change_query_keys": {
            "is_pinned": "options.is_pinned",
            "is_popup": "options.is_popup"
        },
        "indexes": [
            "board_type",
            "category",
            "writer",
            "resource_group",
            "domain_id",
            "user_id",
        ],
    }
